/*-------------------------------------------------*\
 * BobBox.c - Main gui for ORAD Production panel   *
 * was mainz.c - Main gui                          *
 * build: gcc BobBox.c -o BobBox                   *
\*-------------------------------------------------*/
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/time.h>

#define DEBUG        0
#define DEBUG_PADS   0
#define DEBUG_LOGIC  0

#ifndef BOBNUMPADS
#define BOBNUMPADS 16
#endif

/*-------------------------------------------------*/

typedef struct {
	POINT pos;
	int selected;
	} BOBBUTPOS;

BOBBUTPOS bbutpos[BOBNUMPADS][2];

char glText[1024];	// Message text for window
char serdev[255];	// Serial device name

long posX = 100;
long posY = 100;

int mode = 0;          // Which mode are we in !mode=run
int pad_current = -1;  // Which pad pos to be updated
int pad_state = 0;     // pad ON(=1) or OFF(=0)
int done=0;            // Are we finished yet?

unsigned char last_data[256];
unsigned char data[256];
struct termios oldtio,newtio;

HWND hWnd;



/*-------------------------------------------------*/
/*-  Prototypes   ---------------------------------*/

LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);
   void bobUpdateWindow( HWND );
   void DoKeys( WPARAM, LPARAM, int );
   void initpos( void );
   void learnPos( int , int );
   void triggerpad( int , int );
    int openComport( char * );
	int doData( int );
   void parseCmdLine( LPSTR );

/*-------------------------------------------------*/
/*-------------------------------------------------*/


int APIENTRY WinMain(HINSTANCE hInstance,
	HINSTANCE hPrevInstance,
	LPSTR lpCmdLine,
	int nCmdShow)
{
BOOL bRet;
 MSG msg;
 int fd;


	parseCmdLine( lpCmdLine );
	initpos();

	printf("Going to open  %s....", serdev);
	fflush(stdout);
	fd=openComport(serdev);

	memset(data,0,256);       // Clean slate
	memset(last_data,0,256);  // Clean slate

	/*---------  windows stuff ------------*/
	WNDCLASSEX wcex;
	wcex.cbSize = sizeof(wcex);
	wcex.style = CS_HREDRAW | CS_VREDRAW;
	wcex.lpfnWndProc = WndProc;
	wcex.cbClsExtra = 0;
	wcex.cbWndExtra = 0;
	wcex.hInstance = hInstance;
	wcex.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
	wcex.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
	wcex.lpszMenuName = NULL;
	wcex.lpszClassName = "BobBox";
	wcex.hIconSm = NULL;
	if (!RegisterClassEx(&wcex))
		return FALSE;
	hWnd = CreateWindow("BobBox", "BobBox", WS_OVERLAPPEDWINDOW,
		10,10,400,400,
		NULL, NULL, hInstance, NULL);
	if (!hWnd)
		return FALSE;
	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);

	while(!done) {
		if(PeekMessage(&msg, hWnd, 0,0,PM_NOREMOVE)) { 
			bRet=GetMessage(&msg, NULL, 0, 0);
			switch(bRet) {
				case -1: 
					done = 1;
					break;
				case 0:          // No message? WM_QUIT
					done = 1;
				default:        // Process the message...
					TranslateMessage(&msg);
					DispatchMessage(&msg);
					break;
				}
			}
		if( doData(fd) < 1 )   // Only in run mode
			done = 1;
		}
	tcsetattr(fd,TCSANOW,&oldtio);
	close(fd);
	return msg.wParam;
}


void parseCmdLine( LPSTR cmdline )
// This function configures which serial device to open.
{
char val[256];
int port;

	if(!strlen(cmdline)) {  // nothing on command line
		sprintf(serdev,"/dev/ttyS0");  // set to default
		}
	else if(!strncmp(cmdline,"com",3)) {   // are the first 3 letters 'com'?
		sscanf(cmdline,"com%s",val);
		port=atoi(val);
		sprintf(serdev,"/dev/ttyS%d",(port - 1));
		}
	else if (!strncmp(cmdline,"Com",3)) {   // are the first 3 letters 'Com'?
		sscanf(cmdline,"Com%s",val);
		port=atoi(val);
		sprintf(serdev,"/dev/ttyS%d",(port - 1));
		}
	else if (!strncmp(cmdline,"COM",3)) {   // are the first 3 letters 'COM'?
		sscanf(cmdline,"COM%s",val);
		port=atoi(val);
		sprintf(serdev,"/dev/ttyS%d",(port - 1));
		}
	else if (!strncmp(cmdline,"/dev/ttyS",9)) {   // are the first 9 letters '/dev/ttyS'?
		sprintf(serdev,"%s",cmdline);
		}
	else if (!strncmp(cmdline,"ttyS",4)) {   // are the first 4 letters 'ttyS'?
		sprintf(serdev,"/dev/%s",cmdline);
		}
	else
		sprintf(serdev,"/dev/ttyS0");	// set to default

}


void DoKeys( WPARAM wParam, LPARAM lParam, int state)
{
          int nVirtKey; 	// virtual-key code 
unsigned long lKeyData;		// key data 

	nVirtKey = (int) wParam;  	  // virtual-key code 
	lKeyData = (unsigned long)lParam; // key data 

#if DEBUG
	printf("KeyCode: %u   :  %lu : %s\n",nVirtKey,lKeyData,(state?"DOWN":"UP"));
#endif

	switch (nVirtKey) {
		case 76:  // L Key   Toggle Learning ON/OFF
			mode = (!mode);
			learnPos(mode, 0);
		case 78:  // N Key   Skip current button with no change
			if(mode)
				learnPos(mode, -1);
			break;
		case 13:  // Return Key  Record position for current pad
			if(mode)
				learnPos(mode, 0);
			break;
		case 32:  // Spacebar    Reset Value
			if(mode)
				learnPos(mode, 1);
			break;
		case 37:  // leftarrow
			if(mode)
				posX-=10;
			break;
		case 38:  // uparrow
			if(mode)
				posY-=10;
			break;
		case 39:  // rightarrow
			if(mode)
				posX+=10;
			break;
		case 40:  // Downarrow
			if(mode)
				posY+=10;
			break;
		default:
			break;
		}

}


LRESULT CALLBACK WndProc(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch (message) {
		case WM_PAINT:
			bobUpdateWindow(hWnd);
			break;
		case WM_DESTROY:
			done=1;
			PostQuitMessage(0);
			break;
		case WM_KEYDOWN:
			DoKeys(wParam, lParam, 1);
			break;
		case WM_KEYUP:
			// DoKeys(wParam, lParam, 0);
			break;
		default:
			return DefWindowProc(hWnd, message, wParam, lParam);
		}
	return 0;
}


void bobUpdateWindow(HWND hWnd)
{
PAINTSTRUCT  ps;
        HDC  hdc;
       RECT  rt;
        int  height;
        int  i;
       char  binfo[256];

	hdc = BeginPaint(hWnd, &ps);
	GetClientRect(hWnd, &rt);
	height= DrawText(hdc, glText, strlen(glText), &rt, DT_TOP | DT_LEFT);
	rt.top = rt.top+height;
	for(i=0; i < BOBNUMPADS; i++) {
		sprintf(binfo,"PAD %.2d %s ON : %6d, %6d     %s OFF : %6d, %6d",(i+1),
			(bbutpos[i][1].selected)?"**":"  ",
			bbutpos[i][1].pos.x,bbutpos[i][1].pos.y,
			(bbutpos[i][0].selected)?"**":"  ",
			bbutpos[i][0].pos.x,bbutpos[i][0].pos.y);
			
		height= DrawText(hdc, binfo, strlen(binfo), &rt, DT_TOP | DT_LEFT);
		rt.top = rt.top+height;
		}

	EndPaint(hWnd, &ps);
}



void initpos ( void )
/* initialize buttons history */
{
int i;

	for(i=0; i < BOBNUMPADS; i++) {
		bbutpos[i][0].pos.x    = -1;
		bbutpos[i][0].pos.y    = -1;
		bbutpos[i][0].selected = 1;
		bbutpos[i][1].pos.x    = -1;
		bbutpos[i][1].pos.y    = -1;
		bbutpos[i][1].selected = 0;
		}

	mode=1;
	learnPos(mode, 0);
}


void learnPos( int lmode, int skip )
/* Save the coordinates of the but positions */
{
static int learn=0;
POINT curPos;


	switch(lmode) {
		case 0:    // Learn mode off
			if(learn) {  // We were on, turn off
#if DEBUG
				printf("ENDING LEARNING!\n");
#endif
				learn=0;
				sprintf(glText,"RUN MODE\n Press 'L' key for learn mode\n\n");
				}
			break;
		case 1:
			if(learn) { // We are already in learn mode... so learn!
#if DEBUG
				printf("LEARNING!  %d  %s\n",
				        pad_current, (pad_state?"ON":"OFF"));
#endif
				if(skip == 1) {   // set default (-1) values
					bbutpos[pad_current][pad_state].pos.x= -1;
					bbutpos[pad_current][pad_state].pos.y= -1;
					}
				else if(skip == 0) {
					GetCursorPos(&curPos);
					bbutpos[pad_current][pad_state].pos.x=curPos.x;
					bbutpos[pad_current][pad_state].pos.y=curPos.y;
					}
				//  Next pad please 
				if(!pad_state) {
					pad_current++;
					}
				pad_state= (!pad_state);
				if(pad_current>=BOBNUMPADS) {  // Done learning
					learn=0;
					mode =0;
#if DEBUG
					printf("ENDING LEARNING!\n");
#endif
					}
				}
			else {
#if DEBUG
				printf("STARTING LEARNING!\n");
#endif
				learn=1;
				pad_current=0;
				pad_state=1;
				}
			if(learn) {
				sprintf(glText,"Please move mouse to Pad %d %s position\nPress <ENTER> to record,  Press <SPACE> to erase\nPress 'N' to skip\nPress 'L' key to stop learning\n",
				        pad_current+1, (pad_state?"ON":"OFF"));
				}
			else {
				sprintf(glText,"RUN MODE\n Press 'L' key for learn mode\n\n");
				}
			break;
		default:
			fprintf(stderr,"HUH? What mode ar you in?? \n");
			break;
		}

	RedrawWindow(hWnd,NULL,NULL,RDW_INVALIDATE|RDW_ERASE|RDW_ERASENOW);


}


void triggerpad( int pad, int state )
/* Trigger mouse button keypress */
{
POINT curPos;
  int posX;
  int posY;


	posX = bbutpos[pad][state].pos.x;
	posY = bbutpos[pad][state].pos.y;
	if(posX  < 0 )
		return;

	GetCursorPos(&curPos);

	SetCursorPos(posX, posY);
	mouse_event(MOUSEEVENTF_LEFTDOWN ,posX, posY,0,0);
	mouse_event(MOUSEEVENTF_LEFTUP ,posX, posY,0,0);

	mouse_event(MOUSEEVENTF_ABSOLUTE ,curPos.x, curPos.y,0,0);
	SetCursorPos(curPos.x, curPos.y);

}


int openComport( char *name )
{
int fd;

#if DEBUG_PADS
	printf("Going to open  %s....", name);
#endif

	fd=open(name,O_RDWR|O_NOCTTY);
	if(fd < 0) {
		perror("OPPSSS... open failed");
		fprintf(stderr,"Can NOT open:  %s\n", name);
		exit(1);
		}
#if DEBUG_PADS
	printf("OK    fd = %d\n",fd);
#endif

	tcgetattr(fd,&oldtio); /* save current serial port settings */
	bzero(&newtio, sizeof(newtio)); /* clear struct for new port settings */
	newtio.c_cflag = B9600 | CRTSCTS | CS8 | CLOCAL | CREAD; 
	newtio.c_iflag = IGNPAR | ICRNL; /* Raw output. */
	newtio.c_oflag = 0;
	newtio.c_lflag = 0;

	tcflush(fd, TCIFLUSH);
	tcsetattr(fd,TCSANOW,&newtio);

	return(fd);

}


void compareData( void )
/* Compare and send any changes to the screen */
/* Remember that the pads # are bitinverse,   */
/* Byte1,Bit8 = pad1 Byte1,Bit7 = pad2, etc.. */
{
int i;     // Byte index
int j;     // Bit index
int mask;  // Mask
int pad;   // Which pad
int changed;
	
	changed = 0;    // Signal that something is different
	for(i=1; i <= (BOBNUMPADS/8); i++) {
		if(data[i] != last_data[i]) {   // Something changed
			changed = 1;
			mask=128; // start from the beginning...
			for(j=0; j<8; j++) {  // For each bit
				if((data[i]&mask)!=(last_data[i]&mask)) { // pad change
					pad=(i-1)*8+j;
#if DEBUG_LOGIC
					printf("Triggering  PAD %d   STATE %d\n",pad,((data[i]&mask)&&1));
#endif
					triggerpad(pad,((data[i]&mask)&&1));
					bbutpos[pad][((data[i]&mask)&&1)].selected = 1;
					bbutpos[pad][!((data[i]&mask)&&1)].selected = 0;
					}
				mask=mask/2;   // next bit
				}
			last_data[i] = data[i];   //  Remember currnet state
			}
		}

	if(changed) {
#if DEBUG_LOGIC
		printf("Changed!\n");
#endif
		RedrawWindow(hWnd,NULL,NULL,RDW_INVALIDATE|RDW_ERASE|RDW_ERASENOW);
		}
}


int doData( int fd )
/* Get the information from the interface */
{
int want, found, j;
int result, last_byte;
char ans='b';
#if DEBUG_PADS
int i;
#endif

		want=0;
		found=0;
#if DEBUG_PADS
		printf("Going to Write...");
#endif
		result=write(fd,&ans,1);            // Send start command
		if(result < 0) {
			printf("OPPSS  ERROR in read %d : %s\n",errno,strerror(errno));
			return( -1 );
			}
		if(result == 0) {
			printf(" hmmm....  no write\n");
			return( -1 );
			}
#if DEBUG_PADS
		printf("OK,  NOW Going to read\n");
#endif
		while(!want || (want >= found)) {
			result=read(fd,data+found,1);
			if(result < 0) {
				printf("OPPSS  ERROR in read %d : %s\n",errno,strerror(errno));
				return( -1 );
				}
#if DEBUG_PADS
			printf("read result %d : found %d\n",result,found);
			for(i=found; i<found+result ; i++) printf("%4X  ",data[i]);
			printf("\n");
#endif
			if ((!want)&&(data[found] & 0x80)) {
				want = (data[found] & 0x7F);
#if DEBUG_PADS
				printf("Want %d  bytes  (%X)\n",want,data[found]);
#endif
				}
			found += result;		
			}
		/* OK, we have Data!!!  CONVERT IT BACK!!! */
		last_byte=want - ((want/7)+((want%7)&&1));

		for(j=1; j<=last_byte; j++) {
			data[j] |= ((data[last_byte +1+(j/8)] << ((j%8)+(j/8)))&0x80);
			/** Data is inverted, so invert it back.... **/
			data[j]=~data[j];
#if DEBUG_PADS || DEBUG
			printf("%2X  ",data[j]);
#endif
			}

#if DEBUG_PADS || DEBUG
		printf("\n");
#endif
	compareData();

	return(1);
}

/* vim:set sw=4 ts=4: */
